// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "@openzeppelin/contracts/token/ERC721/IERC721.sol";
import "@openzeppelin/contracts/token/ERC1155/IERC1155.sol";
import "@openzeppelin/contracts/metatx/ERC2771Context.sol";
import "@openzeppelin/contracts/token/ERC721/utils/ERC721Holder.sol";
import "@openzeppelin/contracts/token/ERC1155/utils/ERC1155Holder.sol";
import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";
import "@openzeppelin/contracts/utils/cryptography/EIP712.sol";
import "../abstract/roles/EeseeRoleHandler.sol";
import "../interfaces/IEeseeMinter.sol";
import "../interfaces/IEeseeStaking.sol";
import "../interfaces/IEeseeProxy.sol";
import "../types/Call.sol";
import "../abstract/BlastPointsReceiver.sol";
import "../types/Asset.sol";

///@dev Contract for offchain eesee lots logic.
contract EeseeOffchain is ERC2771Context, ERC721Holder, ERC1155Holder, EIP712, EeseeRoleHandler, BlastPointsReceiver {
    using SafeERC20 for IERC20;

    struct SignatureData{
        uint256 nonce;
        uint256 deadline;
        bytes signature;
    }

    ///@dev Is given nonce used.
    mapping(uint256 => bool) public nonceUsed;

    ///@dev ESE token this contract uses.
    IERC20 public immutable ESE;
    ///@dev Eesee staking contract. Tracks volume for this contract.
    IEeseeStaking public immutable staking;
    ///@dev Contract used to call external functions in another context.
    IEeseeProxy private immutable proxy;
    ///@dev Contract that mints NFTs.
    IEeseeMinter private immutable minter;
    ///@dev Signer role af defined in {accessManager}.
    bytes32 private constant SIGNER_ROLE = keccak256("SIGNER_ROLE"); 

    ///@dev Lot typehash.
    bytes32 private constant CREATE_LOT_TYPEHASH = keccak256("CreateLot(bytes32 assetHash,bytes32 dataHash,uint256 nonce,uint256 deadline)");
    ///@dev Ticket typehash.
    bytes32 private constant BUY_TICKETS_TYPEHASH = keccak256("BuyTickets(bytes32 ID,uint256 amount,address recipient,uint256 nonce,uint256 deadline)");
    ///@dev Asset claim typehash.
    bytes32 private constant CLAIM_ASSET_TYPEHASH = keccak256("ClaimAsset(bytes32 ID,bytes32 assetHash,address sender,uint256 nonce,uint256 deadline)");
    ///@dev Tokens claim typehash.
    bytes32 private constant CLAIM_TOKENS_TYPEHASH = keccak256("ClaimTokens(bytes32 ID,uint256 amount,address sender,uint256 nonce,uint256 deadline)");

    event CreateLot(bytes32 indexed ID, address indexed sender, Asset asset, bytes data, uint256 indexed nonce);
    event BuyTickets(bytes32 indexed ID, address sender, address indexed recipient, uint256 tokensSpent, uint256 indexed nonce);

    event ClaimAsset(bytes32 indexed ID, address indexed sender, address recipient, Asset asset, uint256 indexed nonce);
    event ClaimTokens(bytes32 indexed ID, address indexed sender, address recipient, uint256 tokensClaimed, uint256 indexed nonce);

    event AddVolumeReverted(uint256 untrackedVolume, address indexed recipient);
    event RevokeSignature(uint256 indexed nonce);

    error InvalidRecipient();
    error InvalidConstructor();
    error InvalidArrayLengths();
    error ExpiredDeadline();
    error NonceUsed();
    error InvalidAmount();
    error InvalidInterface();
    error InvalidSignature();
    error InvalidValue();
    error TransferNotSuccessful();
    
    modifier isValidRecipient(address recipient){
        if(recipient == address(0)) revert InvalidRecipient();
        _;
    }

    constructor(
        IERC20 _ESE,
        IEeseeStaking _staking,
        IEeseeProxy _proxy,
        IEeseeMinter _minter,
        IEeseeAccessManager _accessManager,

        IBlastPoints _blastPoints,
        address _blastPointsOperator,

        address trustedForwarder
    ) ERC2771Context(trustedForwarder) BlastPointsReceiver(_blastPoints, _blastPointsOperator) EIP712("EeseeOffchain", "1") EeseeRoleHandler(_accessManager) {
        if(
            address(_ESE) == address(0) || 
            address(_proxy) == address(0) ||
            address(_minter) == address(0)
        ) revert InvalidConstructor();

        ESE = _ESE;
        staking = _staking;
        proxy = _proxy;
        minter = _minter;
    }

    // ============ Public Write Functions ============

    /**
     * @dev Creates lots with offchain logic. Emits {CreateLot} events for each created lot.
     * @param assets - Assets to list. Note: The sender must have them approved for this contract.
     * @param data - Lot data for offchain logic.
     * @param signatureData - Signatures for asset and data signed by SIGNER_ROLE.

     * @return IDs - IDs of lots created.
     */
    function createLots(
        Asset[] calldata assets, 
        bytes[] calldata data,
        SignatureData[] calldata signatureData
    ) external payable returns(bytes32[] memory IDs){
        if(assets.length != data.length || assets.length != signatureData.length) revert InvalidArrayLengths();
        IDs = new bytes32[](assets.length);

        uint256 valueLeft = msg.value;
        address msgSender = _msgSender();
        for(uint256 i; i < assets.length;){
            IDs[i] = getCreateLotStructHash(assets[i], data[i], signatureData[i].nonce, signatureData[i].deadline);
            _checkSignature(
                msgSender, 
                IDs[i], 
                signatureData[i]
            );
            valueLeft = _transferAssetFrom(assets[i], msgSender, valueLeft);

            emit CreateLot(IDs[i], msgSender, assets[i], data[i], signatureData[i].nonce);
            unchecked{ ++i; }
        }
        if(valueLeft != 0) revert InvalidValue();
    }

    /**
     * @dev Buys tickets to participate in a lot. Emits {BuyTicket} event for lot the ticket was bought.
     * @param IDs - IDs of lots to buy tickets for.
     * @param tokenAmounts - ESE amount to spend in each lot.
     * @param signatureData - Signatures for ID, tokenAmount and recipient signed by SIGNER_ROLE.
     * @param recipient - Recipient of tickets.
     * @param permit - Abi-encoded ESE permit data containing approveAmount, deadline, v, r and s. Set to empty bytes to skip permit.

     * @return tokensSpent - ESE tokens spent.
     */
    function buyTickets(
        bytes32[] calldata IDs,
        uint256[] calldata tokenAmounts, 
        SignatureData[] calldata signatureData,
        address recipient,
        bytes calldata permit
    ) isValidRecipient(recipient) external returns(uint256 tokensSpent){
        if(IDs.length != tokenAmounts.length || IDs.length != signatureData.length) revert InvalidArrayLengths();

        address msgSender = _msgSender();
        for(uint256 i; i < IDs.length;){
            _checkSignature(
                msgSender, 
                getBuyTicketsStructHash(IDs[i], tokenAmounts[i], recipient, signatureData[i].nonce, signatureData[i].deadline), 
                signatureData[i]
            );
            tokensSpent += tokenAmounts[i];

            emit BuyTickets(IDs[i], msgSender, recipient, tokenAmounts[i], signatureData[i].nonce);
            unchecked { ++i; }
        }

        _permit(IERC20Permit(address(ESE)), msgSender, permit);
        if(tokensSpent == 0) return 0;

        ESE.safeTransferFrom(msgSender, address(this), tokensSpent);
        if(address(staking) != address(0)){
            try staking.addVolume(uint96(tokensSpent), recipient) {} catch {
                emit AddVolumeReverted(tokensSpent, recipient);
            }
        }
    }

    /**
     * @dev Claims assets from a lot. Emits {ClaimAssets} event for each asset claimed.
     * @param IDs - IDs of lots to claim assets in.
     * @param assets - Assets to claim and send to recipient.
     * @param signatureData - Signature for ID, asset and msgSender signed by SIGNER_ROLE.
     * @param recipient - Recipient of assets.
     * @param _call - Additional external call to make.

     * @return returnData - Data returned from external call.
     */
    function claimAssets(
        bytes32[] calldata IDs,
        Asset[] calldata assets, 
        SignatureData[] calldata signatureData,
        address recipient,
        Call calldata _call
    ) payable isValidRecipient(recipient) external returns (bytes memory returnData){
        if(IDs.length != assets.length || IDs.length != signatureData.length) revert InvalidArrayLengths();
        
        address msgSender = _msgSender();
        for(uint256 i; i < IDs.length;){
            _checkSignature(
                msgSender, 
                getClaimAssetStructHash(IDs[i], assets[i], msgSender, signatureData[i].nonce, signatureData[i].deadline), 
                signatureData[i]
            );
            _transferAssetTo(assets[i], recipient);

            emit ClaimAsset(IDs[i], msgSender, recipient, assets[i], signatureData[i].nonce);
            unchecked { ++i; }
        }

        if(_call.target != address(0)) returnData = proxy.callExternal{value: msg.value}(_call.target, _call.callData);
    }

    /**
     * @dev Claims ESE tokens from a lot. Emits {ClaimTokens} event for each lot the tokens were claimed from.
     * @param IDs - IDs of lots to claim tokens in.
     * @param tokenAmounts - Amounts of tokens to claim in each lot.
     * @param signatureData - Signature for ID, tokenAmount and msgSender signed by SIGNER_ROLE.
     * @param recipient - Recipient of tokens.
     * @param _call - Additional external call to make.

     * @return tokensClaimed - ESE sent to recipient in this call.
     * @return returnData - Data returned from external call.
     */
    function claimTokens(
        bytes32[] calldata IDs,
        uint256[] calldata tokenAmounts, 
        SignatureData[] calldata signatureData,
        address recipient,
        Call calldata _call
    ) payable isValidRecipient(recipient) external returns (uint256 tokensClaimed, bytes memory returnData){
        if(IDs.length != tokenAmounts.length || IDs.length != signatureData.length) revert InvalidArrayLengths();
        
        address msgSender = _msgSender();
        for(uint256 i; i < IDs.length;){
            _checkSignature(
                msgSender, 
                getClaimTokensStructHash(IDs[i], tokenAmounts[i], msgSender, signatureData[i].nonce, signatureData[i].deadline), 
                signatureData[i]
            );
            tokensClaimed += tokenAmounts[i];

            emit ClaimTokens(IDs[i], msgSender, recipient, tokenAmounts[i], signatureData[i].nonce);
            unchecked { ++i; }
        }
        ESE.safeTransfer(recipient, tokensClaimed);

        if(_call.target != address(0)) returnData = proxy.callExternal{value: msg.value}(_call.target, _call.callData);
    }

    /**
     * @dev Callable by SIGNER_ROLE to revoke signatures. Emits {RevokeSignature} for each signature revoked.
     * @param nonces - Signature nonces to revoke.
     */
    function revokeSignatures(uint256[] calldata nonces) external {
        if(!isSigner(_msgSender())) revert CallerNotAuthorized();
        for(uint256 i; i < nonces.length;){
            if(nonceUsed[nonces[i]] == false){
                nonceUsed[nonces[i]] = true;
                emit RevokeSignature(nonces[i]);
            }
            unchecked { ++i; }
        }
    }

    // ============ Public View Functions ============

    function getCreateLotStructHash(Asset calldata asset, bytes calldata data, uint256 nonce, uint256 deadline) public pure returns (bytes32){
        return keccak256(abi.encode(
            CREATE_LOT_TYPEHASH, 
            keccak256(abi.encode(asset)),
            keccak256(data),
            nonce,
            deadline
        ));
    }

    function getBuyTicketsStructHash(bytes32 ID, uint256 amount, address recipient, uint256 nonce, uint256 deadline) public pure returns (bytes32){
        return keccak256(abi.encode(
            BUY_TICKETS_TYPEHASH, 
            ID,
            amount,
            recipient,
            nonce,
            deadline
        ));
    }

    function getClaimAssetStructHash(bytes32 ID, Asset calldata asset, address sender, uint256 nonce, uint256 deadline) public pure returns (bytes32){
        return keccak256(abi.encode(
            CLAIM_ASSET_TYPEHASH, 
            ID,
            keccak256(abi.encode(asset)),
            sender,
            nonce,
            deadline
        ));
    }

    function getClaimTokensStructHash(bytes32 ID, uint256 amount, address sender, uint256 nonce, uint256 deadline) public pure returns (bytes32){
        return keccak256(abi.encode(
            CLAIM_TOKENS_TYPEHASH, 
            ID,
            amount,
            sender,
            nonce,
            deadline
        ));
    }

    // ============ Internal Write Functions ============

    function _checkSignature(
        address msgSender,
        bytes32 structHash,
        SignatureData calldata signatureData
    ) internal {
        if(block.timestamp > signatureData.deadline) revert ExpiredDeadline();
        if(nonceUsed[signatureData.nonce]) revert NonceUsed();
        nonceUsed[signatureData.nonce] = true;
        
        if(isSigner(msgSender)) return;

        address signer = ECDSA.recover(_hashTypedDataV4(structHash), signatureData.signature);
        if(!isSigner(signer)) revert InvalidSignature();
    }
    
    function _transferAssetFrom(
        Asset calldata asset,
        address from,
        uint256 availableValue
    ) internal returns(uint256 valueLeft) {
        valueLeft = availableValue;
        if (asset.assetType == AssetType.ERC721) {
            if(asset.amount != 1) revert InvalidAmount();
            (bool success, bytes memory res) = asset.token.staticcall(abi.encodeWithSelector(IERC165.supportsInterface.selector, type(IERC721).interfaceId));
            if(!success || abi.decode(res, (bool)) == false) revert InvalidInterface();

            IERC721(asset.token).safeTransferFrom(from, address(this), asset.tokenID);
        } else if (asset.assetType == AssetType.ERC1155) {
            if(asset.amount == 0) revert InvalidAmount();
            (bool success, bytes memory res) = asset.token.staticcall(abi.encodeWithSelector(IERC165.supportsInterface.selector, type(IERC1155).interfaceId));
            if(!success || abi.decode(res, (bool)) == false) revert InvalidInterface();

            IERC1155(asset.token).safeTransferFrom(from, address(this), asset.tokenID, asset.amount, "");
        } else if (asset.assetType == AssetType.ERC20) {
            if(asset.amount == 0) revert InvalidAmount();
            // Most ERC-20 tokens do not implement supportsInterface, so we check against IERC721 to ensure we are not calling ERC721 contract.
            // ERC1155 will revert anyway because of different safeTransferFrom scheme.
            (bool success, bytes memory res) = asset.token.staticcall(abi.encodeWithSelector(IERC165.supportsInterface.selector, type(IERC721).interfaceId));
            if(success && abi.decode(res, (bool)) == true) revert InvalidInterface();
            _permit(IERC20Permit(asset.token), from, asset.data);

            IERC20(asset.token).safeTransferFrom(from, address(this), asset.amount);
        } else if (asset.assetType == AssetType.Native) {
            if(asset.amount == 0) revert InvalidAmount();
            if(availableValue < asset.amount) revert InvalidValue();
            unchecked{ valueLeft -= asset.amount; }
        } else if (asset.assetType == AssetType.ESE) {

        } else /*if (asset.assetType == AssetType.ERC721LazyMint)*/ {
            if(asset.amount != 1) revert InvalidAmount();
        }
    }

    function _transferAssetTo(
        Asset calldata asset,
        address to
    ) internal {
        if (asset.assetType == AssetType.ERC721) {
            IERC721(asset.token).safeTransferFrom(address(this), to, asset.tokenID, asset.data);
        } else if (asset.assetType == AssetType.ERC1155) {
            IERC1155(asset.token).safeTransferFrom(address(this), to, asset.tokenID, asset.amount, asset.data);
        } else if (asset.assetType == AssetType.ERC20) {
            IERC20(asset.token).safeTransfer(to, asset.amount);
        } else if (asset.assetType == AssetType.Native) {
            (bool success,) = to.call{value: asset.amount}("");
            if(!success) revert TransferNotSuccessful(); 
        } else if(asset.assetType == AssetType.ESE) {
            ESE.safeTransfer(to, asset.amount);
        } else /*if(asset.assetType == AssetType.ERC721LazyMint)*/ {
            (
                uint256 collectionID, 
                address owner,
                LazyMintCollectionMetadata memory collectionMetadata,
                LazyMintTokenMetadata memory tokenMetadata
            ) = abi.decode(asset.data, (uint256, address, LazyMintCollectionMetadata, LazyMintTokenMetadata));
            minter.lazyMint(collectionID, owner, collectionMetadata, tokenMetadata, to);
        }
    }

    function _permit(IERC20Permit token, address spender, bytes calldata permit) internal {
        if(permit.length != 0){
            (uint256 approveAmount, uint256 deadline, uint8 v, bytes32 r, bytes32 s) = abi.decode(permit, (uint256, uint256, uint8, bytes32, bytes32));
            token.permit(spender, address(this), approveAmount, deadline, v, r, s);
        }
    }

    // ============ Internal View Functions ============

    function isSigner(address _addr) internal view returns(bool){
        return _hasRole(SIGNER_ROLE, _addr);
    }

    function _msgSender() internal view override(Context, ERC2771Context) returns (address) {
        return ERC2771Context._msgSender();
    }

    function _msgData() internal view override(Context, ERC2771Context) returns (bytes calldata) {
        return ERC2771Context._msgData();
    }

    function _contextSuffixLength() internal view override(Context, ERC2771Context) returns (uint256) {
        return ERC2771Context._contextSuffixLength();
    }
}